from cepimose.enums import AgeGroup, Manufacturer, Region
import unittest
import cepimose
import datetime
from nose.plugins.attrib import attr


class CepimoseTestCase(unittest.TestCase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def assertDatesIncreaseSince(self, data, startDate):
        previousDate = startDate - datetime.timedelta(days=1)
        for row in data:
            self.assertGreater(row.date, previousDate, row)
            previousDate = row.date

    @attr("sledilnik")
    def test_vaccinations_by_day(self):
        # Test feature one.
        data = cepimose.vaccinations_by_day()
        self.assertGreater(len(data), 150)

        def assertRow(row, expected_date, expected_first, expected_second):
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.first_dose, expected_first, delta=400)
            self.assertAlmostEqual(row.second_dose, expected_second, delta=400)

        #! NIJZ is changing data tests could fail in the future
        assertRow(data[9], datetime.datetime(2021, 1, 5), 15711, 0)
        assertRow(data[22], datetime.datetime(2021, 1, 18), 49100, 315)
        assertRow(data[41], datetime.datetime(2021, 2, 6), 56066, 44924)
        assertRow(data[42], datetime.datetime(2021, 2, 7), 56066, 44924)

        # values should be growing
        firstPrevious = 0
        secondPrevious = 0
        for row in data:
            print(row, firstPrevious, secondPrevious)
            self.assertGreaterEqual(row.first_dose, firstPrevious)
            self.assertGreaterEqual(row.second_dose, secondPrevious)
            firstPrevious = row.first_dose
            secondPrevious = row.second_dose

        self.assertDatesIncreaseSince(data, datetime.datetime(2020, 12, 27))

    def test_vaccinations_by_age(self):
        # Test feature one.
        data = {row.age_group: row for row in cepimose.vaccinations_by_age()}

        expected_age_groups = [
            "0-17",
            "18-24",
            "25-29",
            "30-34",
            "35-39",
            "40-44",
            "45-49",
            "50-54",
            "55-59",
            "60-64",
            "65-69",
            "70-74",
            "75-79",
            "80-84",
            "85-89",
            "90+",
        ]

        self.assertTrue(len(data), len(expected_age_groups))

        for grp in expected_age_groups:
            self.assertGreater(data[grp].count_first, 0)
            self.assertGreater(data[grp].share_first, 0)
            self.assertGreater(data[grp].count_second, 0)
            self.assertGreater(data[grp].share_second, 0)
            self.assertGreaterEqual(data[grp].count_first, data[grp].count_second)

    @attr("sledilnik")
    def test_vaccines_supplied_by_manufacturer(self):
        data = cepimose.vaccines_supplied_by_manufacturer()
        self.assertTrue(len(data) > 50)

        def assertRow(row, expected_date, expected):
            self.assertEqual(row.date, expected_date)
            self.assertEqual(row.pfizer, expected[0])
            self.assertEqual(row.moderna, expected[1])
            self.assertEqual(row.az, expected[2])
            self.assertEqual(row.janssen, expected[3])

        assertRow(
            data[1], datetime.datetime(2020, 12, 30), [8190, None, None, None]
        )  # R = 2
        assertRow(
            data[3], datetime.datetime(2021, 1, 11), [19890, None, None, None]
        )  # R = 6
        assertRow(
            data[16], datetime.datetime(2021, 2, 25), [None, 8400, 16800, None]
        )  # combined: two response data items with same date; second has R = 1
        assertRow(data[32], datetime.datetime(2021, 4, 14), [None, None, None, 7200])

        self.assertDatesIncreaseSince(data, datetime.datetime(2020, 12, 26))

    @attr("sledilnik")
    def test_vaccinations_by_region_by_day(self):
        data = cepimose.vaccinations_by_region_by_day()
        expected_keys = [key for key in cepimose.enums.Region]

        self.assertEquals(expected_keys, list(data.keys()), "Object keys")

        for element in data.items():
            print(element[0])
            self.assertTrue(len(element[1]) != 0)
            self.assertDatesIncreaseSince(element[1], datetime.datetime(2020, 12, 27))

            # values should be growing
            firstPrevious = 0
            secondPrevious = 0
            for row in data[element[0]]:
                print(row, firstPrevious, secondPrevious)
                self.assertGreaterEqual(row.first_dose, firstPrevious)
                self.assertGreaterEqual(row.second_dose, secondPrevious)
                firstPrevious = row.first_dose
                secondPrevious = row.second_dose

        pomurska_region = data[cepimose.data.Region.POMURSKA]

        def assertRow(row, expected_date, expected_first, expected_second):
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.first_dose, expected_first, delta=30)
            self.assertAlmostEqual(row.second_dose, expected_second, delta=30)

        assertRow(pomurska_region[9], datetime.datetime(2021, 1, 5), 988, 0)
        assertRow(pomurska_region[22], datetime.datetime(2021, 1, 18), 2847, 5)

    def test_vaccinations_by_region_by_day_with_arg(self):
        data = cepimose.vaccinations_by_region_by_day(cepimose.data.Region.POMURSKA)
        expected_keys = [cepimose.data.Region.POMURSKA]
        self.assertEquals(expected_keys, list(data.keys()), "Object keys")

        pomurska_region = data[cepimose.data.Region.POMURSKA]

        def assertRow(row, expected_date, expected_first, expected_second):
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.first_dose, expected_first, delta=30)
            self.assertAlmostEqual(row.second_dose, expected_second, delta=30)

        assertRow(pomurska_region[9], datetime.datetime(2021, 1, 5), 988, 0)
        assertRow(pomurska_region[22], datetime.datetime(2021, 1, 18), 2847, 5)

    @attr("sledilnik")
    def test_vaccinations_by_municipalities_share(self):
        data = cepimose.vaccinations_by_municipalities_share()

        mun_number = 212

        self.assertEqual(len(data), mun_number, "Unexpected number of municipalities")

        for m in data:
            print(m)
            self.assertGreater(m.dose1, 0)
            self.assertGreater(m.dose2, 0)
            self.assertGreaterEqual(m.dose1, m.dose2)
            self.assertGreaterEqual(m.population, m.dose1)
            self.assertGreaterEqual(m.population, m.dose2)

            self.assertGreater(m.share1, 0)
            self.assertGreater(m.share2, 0)
            self.assertGreaterEqual(m.share1, m.share2)
            self.assertGreaterEqual(1, m.share1)
            self.assertGreaterEqual(1, m.share2)

    def test_vaccination_timestamp(self):
        ts = cepimose.vaccinations_timestamp()
        print("Last update:", ts)
        day_delta = datetime.timedelta(days=3)
        today = datetime.datetime.now()
        diff = today - ts
        self.assertGreaterEqual(day_delta, diff)
        # self.assertGreaterEqual(today, ts) // TODO: adjust timezone for github actions

    def test_vaccinations_by_manufacturer_supplied_used(self):
        data = cepimose.vaccinations_by_manufacturer_supplied_used()
        expected_keys = [key for key in cepimose.enums.Manufacturer]

        self.assertEquals(expected_keys, list(data.keys()), "Dict keys")

        Test_data = {
            Manufacturer.JANSSEN: {
                "row": 1,
                "date": datetime.datetime(2021, 4, 30),
                "supplied": 13200,
                "used": 73,
            },
            Manufacturer.MODERNA: {
                "row": 2,
                "date": datetime.datetime(2021, 2, 5),
                "supplied": 8400,
                "used": 996,
            },
            Manufacturer.PFIZER: {
                "row": 2,
                "date": datetime.datetime(2021, 1, 4),
                "supplied": 39780,
                "used": 13259,
            },
            Manufacturer.AZ: {
                "row": 2,
                "date": datetime.datetime(2021, 2, 18),
                "supplied": 36000,
                "used": 4462,
            },
        }

        def assertRow(row, expected_date, expected_first, expected_second):
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.supplied, expected_first, delta=300)
            self.assertAlmostEqual(row.used, expected_second, delta=300)

        for key, group_data in data.items():
            print(key, len(group_data))
            self.assertTrue(len(group_data) > 0)
            self.assertDatesIncreaseSince(group_data, datetime.datetime(2020, 12, 26))
            test_item = Test_data[key]
            assertRow(
                group_data[test_item["row"]],
                test_item["date"],
                test_item["supplied"],
                test_item["used"],
            )
            # ? more assertions

    @attr("sledilnik")
    def test_vaccinations_by_manufacturer_used(self):
        data = cepimose.vaccinations_by_manufacturer_used()

        self.assertDatesIncreaseSince(data, datetime.datetime(2020, 12, 26))

        def assertRow(row, expected_date, expected):
            print(row)
            expected_pfizer, expected_moderna, expected_az, expected_janssen = expected
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.pfizer, expected_pfizer, delta=50)
            self.assertAlmostEqual(row.moderna, expected_moderna, delta=50)
            self.assertAlmostEqual(row.az, expected_az, delta=50)
            self.assertAlmostEqual(row.janssen, expected_janssen, delta=50)

        assertRow(data[20], datetime.datetime(2021, 1, 16), [323, None, None, None])
        assertRow(data[23], datetime.datetime(2021, 1, 19), [2103, 66, None, None])
        assertRow(data[38], datetime.datetime(2021, 2, 3), [4805, None, 1, None])
        assertRow(data[42], datetime.datetime(2021, 2, 7), [None, None, None, None])
        assertRow(data[50], datetime.datetime(2021, 2, 15), [28, 40, 18, None])
        assertRow(data[79], datetime.datetime(2021, 3, 16), [603, 445, 12, None])
        assertRow(data[98], datetime.datetime(2021, 4, 4), [None, 1594, None, None])
        assertRow(data[99], datetime.datetime(2021, 4, 5), [1, None, None, None])
        assertRow(data[120], datetime.datetime(2021, 4, 26), [1, None, 381, None])
        assertRow(data[134], datetime.datetime(2021, 5, 10), [46, 141, 2080, 717])

        for row in data:
            print(row)
            self.assertTrue(
                row.date >= datetime.datetime(2021, 1, 8, 0, 0) or row.moderna == None,
                f"Too early for Moderna usage: {row}",
            )
            self.assertTrue(
                row.date >= datetime.datetime(2021, 1, 28, 0, 0) or row.az == None,
                f"Too early for Astra Zeneca usage: {row}",
            )
            self.assertTrue(
                row.date >= datetime.datetime(2021, 4, 14, 0, 0) or row.janssen == None,
                f"Too early for Janssen usage: {row}",
            )

    def test_vaccine_supply_and_usage(self):
        data = cepimose.vaccines_supplied_and_used()
        self.assertGreater(len(data), 100)

        def assertRow(row, expected_date, expected_supp, expected_used):
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.supplied, expected_supp, delta=400)
            self.assertAlmostEqual(row.used, expected_used, delta=400)

        #! NIJZ is changing data tests could fail in the future
        assertRow(data[9], datetime.datetime(2021, 1, 4), 39780, 13248)
        assertRow(data[22], datetime.datetime(2021, 1, 17), 60870, 49189)

        self.assertDatesIncreaseSince(data, datetime.datetime(2020, 12, 26))

    def test_vaccinations_by_region(self):
        # Test feature one.
        data = {row.region: row for row in cepimose.vaccinations_by_region()}

        expected_regions = [
            "Koroška",
            "Zasavska",
            "Goriška",
            "Posavska",
            "Gorenjska",
            "Podravska",
            "Pomurska",
            "Osrednjeslovenska",
            "Jugovzhodna Slovenija",
            "Primorsko-notranjska",
            "Savinjska",
            "Obalno-kraška",
        ]

        print(data.keys())

        self.assertTrue(len(data), len(expected_regions))

        for grp in expected_regions:
            self.assertGreater(data[grp].count_first, 0)
            self.assertGreater(data[grp].share_first, 0)
            self.assertGreater(data[grp].count_second, 0)
            self.assertGreater(data[grp].share_second, 0)

    def test_supplied_by_manufacturer_cumulative(self):
        data = cepimose.vaccines_supplied_by_manufacturer_cumulative()
        self.assertTrue(len(data) > 10)

        def assertRow(row, expected_date, expected):
            self.assertEqual(row.date, expected_date)
            self.assertEqual(row.pfizer, expected[0])
            self.assertEqual(row.moderna, expected[1])
            self.assertEqual(row.az, expected[2])
            self.assertEqual(row.janssen, expected[3])

        assertRow(data[3], datetime.datetime(2021, 1, 11), [59670, None, None, None])
        assertRow(data[7], datetime.datetime(2021, 1, 31), [None, 3600, None, None])
        assertRow(data[10], datetime.datetime(2021, 2, 6), [None, None, 9600, None])
        assertRow(data[16], datetime.datetime(2021, 2, 25), [None, 16800, 52800, None])
        assertRow(data[32], datetime.datetime(2021, 4, 14), [None, None, None, 7200])

        self.assertDatesIncreaseSince(data, datetime.datetime(2020, 12, 26))

    @attr("sledilnik")
    def test_vaccinations_by_age_group(self):
        data = cepimose.vaccinations_by_age_group()
        expected_keys = [key for key in cepimose.enums.AgeGroup]

        self.assertEquals(expected_keys, list(data.keys()), "Dict keys")

        for key, group_data in data.items():
            print(key, len(group_data))
            self.assertTrue(len(group_data) != 0)
            self.assertDatesIncreaseSince(group_data, datetime.datetime(2020, 12, 27))

            for group_day in group_data:
                self.assertGreaterEqual(group_day.first_dose, 0)
                self.assertGreaterEqual(group_day.second_dose, 0)

    def test_vaccinations_by_age_group_with_arg(self):
        data = cepimose.vaccinations_by_age_group(cepimose.enums.AgeGroup.GROUP_90)

        self.assertTrue(len(data) > 10)

        def assertRow(row, expected_date, expected_dose):
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.first_dose, expected_dose[0], delta=100)
            self.assertAlmostEqual(row.second_dose, expected_dose[1], delta=100)

        assertRow(data[21], datetime.datetime(2021, 1, 17), [3623, 1])
        assertRow(data[70], datetime.datetime(2021, 3, 7), [7920, 4821])

        self.assertDatesIncreaseSince(data, datetime.datetime(2020, 12, 26))

    def test_vaccinations_age_group_by_region_on_day(self):
        data = cepimose.vaccinations_age_group_by_region_on_day()
        expected_keys = [key for key in cepimose.enums.AgeGroup]

        self.assertEquals(expected_keys, list(data.keys()), "Dict keys")

        for key, group_data in data.items():
            print(key, len(group_data))
            self.assertTrue(len(group_data) == len(list(cepimose.Region)))

            for group_day in group_data:
                print(group_day)
                self.assertEqual(group_day.region, group_day.dose1.region)
                self.assertEqual(group_day.region, group_day.dose2.region)
                self.assertEqual(group_day.dose1.region, group_day.dose2.region)
                self.assertGreaterEqual(group_day.dose1.total_count, 0)
                self.assertGreaterEqual(group_day.dose1.group_count, 0)
                self.assertGreaterEqual(group_day.dose1.total_share, 0)
                self.assertGreaterEqual(group_day.dose1.group_share, 0)
                self.assertGreaterEqual(group_day.dose2.total_count, 0)
                self.assertGreaterEqual(group_day.dose2.group_count, 0)
                self.assertGreaterEqual(group_day.dose2.total_share, 0)
                self.assertGreaterEqual(group_day.dose2.group_share, 0)

    def test_vaccinations_age_group_by_region_on_day_with_arg(self):
        chosen_group = cepimose.AgeGroup.GROUP_0_17
        data = cepimose.vaccinations_age_group_by_region_on_day(chosen_group)

        self.assertTrue(len(data) == len(list(cepimose.Region)))

        region_names = [name.value for name in list(cepimose.Region)]

        for item in data:
            print(item)
            self.assertTrue(item.region in region_names)

            self.assertTrue(item.region, item.dose1.region)
            self.assertGreaterEqual(item.dose1.total_count, 0)
            self.assertGreaterEqual(item.dose1.group_count, 0)
            self.assertGreaterEqual(item.dose1.total_count, item.dose1.group_count)
            self.assertGreaterEqual(item.dose1.total_share, 0)
            self.assertGreaterEqual(item.dose1.group_share, 0)
            self.assertGreaterEqual(item.dose1.total_share, item.dose1.group_share)

            self.assertTrue(item.region, item.dose2.region)
            self.assertGreaterEqual(item.dose2.total_count, 0)
            self.assertGreaterEqual(item.dose2.group_count, 0)
            self.assertGreaterEqual(item.dose2.total_count, item.dose2.group_count)
            self.assertGreaterEqual(item.dose2.total_share, 0)
            self.assertGreaterEqual(item.dose2.group_share, 0)
            self.assertGreaterEqual(item.dose2.total_share, item.dose2.group_share)

    def test_vaccinations_date_range(self):
        start_date = datetime.datetime(2021, 5, 6)
        # assert args end_date and start_date are equal
        data = cepimose.vaccinations_date_range(
            end_date=start_date, start_date=start_date
        )
        print(data)
        self.assertEqual(start_date, data.date_from)
        self.assertEqual(data.date_from, data.date_to)
        self.assertEqual(data.property, None)
        self.assertEqual(len(data.by_day), 1)
        self.assertAlmostEqual(data.male.dose1, 7906, delta=50)
        self.assertAlmostEqual(data.male.dose2, 5791, delta=50)
        self.assertAlmostEqual(data.female.dose1, 7684, delta=50)
        self.assertAlmostEqual(data.female.dose2, 6288, delta=50)
        self.assertEqual(data.by_day[0].first_dose, data.male.dose1 + data.female.dose1)
        self.assertEqual(
            data.by_day[0].second_dose, data.male.dose2 + data.female.dose2
        )
        self.assertAlmostEqual(data.pfizer.dose1, 13945, delta=50)
        self.assertAlmostEqual(data.pfizer.dose2, 11382, delta=50)
        self.assertAlmostEqual(data.az.dose1, 1332, delta=50)
        self.assertAlmostEqual(data.az.dose2, 473, delta=50)
        self.assertAlmostEqual(data.moderna.dose1, 253, delta=50)
        self.assertAlmostEqual(data.moderna.dose2, 164, delta=50)
        self.assertAlmostEqual(data.janssen.dose1, 60, delta=50)
        self.assertAlmostEqual(data.janssen.dose2, 0, delta=50)

        # assert arg start_date is greater than end_date
        end_date = datetime.datetime(2021, 2, 28)
        with self.assertRaises(Exception):
            cepimose.vaccinations_date_range(end_date=end_date, start_date=start_date)

        start_date = datetime.datetime(2020, 12, 26)
        today = datetime.datetime.today()
        end_date = datetime.datetime(today.year, today.month, today.day)
        data = cepimose.vaccinations_date_range(
            end_date=end_date, start_date=start_date
        )
        self.assertDatesIncreaseSince(data.by_day, datetime.datetime(2020, 12, 26))

        self.assertEqual(start_date, data.date_from)
        self.assertEqual(end_date, data.date_to)

    def test_vaccinations_date_range_region(self):
        property = Region.POMURSKA

        start_date = datetime.datetime(2021, 4, 20)
        # assert args end_date and start_date are equal
        data = cepimose.vaccinations_date_range(
            end_date=start_date, start_date=start_date, property=property
        )
        print(data)
        self.assertEqual(start_date, data.date_from)
        self.assertEqual(data.date_from, data.date_to)
        self.assertEqual(data.property, property)
        self.assertEqual(len(data.by_day), 1)
        self.assertAlmostEqual(data.male.dose1, 18, delta=5)
        self.assertAlmostEqual(data.male.dose2, 147, delta=10)
        self.assertAlmostEqual(data.female.dose1, 10, delta=5)
        self.assertAlmostEqual(data.female.dose2, 165, delta=10)
        self.assertEqual(data.by_day[0].first_dose, data.male.dose1 + data.female.dose1)
        self.assertEqual(
            data.by_day[0].second_dose, data.male.dose2 + data.female.dose2
        )
        self.assertAlmostEqual(data.pfizer.dose1, 8, delta=10)
        self.assertAlmostEqual(data.pfizer.dose2, 312, delta=10)
        self.assertAlmostEqual(data.az.dose1, 20, delta=10)
        self.assertAlmostEqual(data.az.dose2, 0, delta=10)
        self.assertAlmostEqual(data.moderna.dose1, 0, delta=10)
        self.assertAlmostEqual(data.moderna.dose2, 0, delta=10)
        self.assertAlmostEqual(data.janssen.dose1, 0, delta=10)
        self.assertAlmostEqual(data.janssen.dose2, 0, delta=10)

        # assert arg start_date is greater than end_date
        end_date = datetime.datetime(2021, 2, 28)
        with self.assertRaises(Exception):
            cepimose.vaccinations_date_range(
                end_date=end_date, start_date=start_date, property=property
            )

        start_date = datetime.datetime(2020, 12, 26)
        today = datetime.datetime.today()
        end_date = datetime.datetime(today.year, today.month, today.day)
        data = cepimose.vaccinations_date_range(
            end_date=end_date, start_date=start_date, property=property
        )
        self.assertDatesIncreaseSince(data.by_day, datetime.datetime(2020, 12, 26))

        self.assertEqual(start_date, data.date_from)
        self.assertEqual(end_date, data.date_to)

    def test_vaccinations_date_range_age_group(self):
        property = AgeGroup.GROUP_70_74

        start_date = datetime.datetime(2021, 4, 20)
        # assert args end_date and start_date are equal
        data = cepimose.vaccinations_date_range(
            end_date=start_date, start_date=start_date, property=property
        )
        self.assertEqual(len(data.by_day), 1)

        print(data)
        self.assertEqual(start_date, data.date_from)
        self.assertEqual(data.date_from, data.date_to)
        self.assertEqual(data.property, property)
        self.assertEqual(len(data.by_day), 1)
        self.assertAlmostEqual(data.male.dose1, 46, delta=5)
        self.assertAlmostEqual(data.male.dose2, 533, delta=10)
        self.assertAlmostEqual(data.female.dose1, 42, delta=5)
        self.assertAlmostEqual(data.female.dose2, 579, delta=10)
        self.assertEqual(data.by_day[0].first_dose, data.male.dose1 + data.female.dose1)
        self.assertEqual(
            data.by_day[0].second_dose, data.male.dose2 + data.female.dose2
        )
        self.assertAlmostEqual(data.pfizer.dose1, 41, delta=10)
        self.assertAlmostEqual(data.pfizer.dose2, 1111, delta=10)
        self.assertAlmostEqual(data.az.dose1, 5, delta=10)
        self.assertAlmostEqual(data.az.dose2, 1, delta=10)
        self.assertAlmostEqual(data.moderna.dose1, 42, delta=10)
        self.assertAlmostEqual(data.moderna.dose2, 0, delta=10)
        self.assertAlmostEqual(data.janssen.dose1, 0, delta=10)
        self.assertAlmostEqual(data.janssen.dose2, 0, delta=10)

        # assert arg start_date is greater than arg end_date
        end_date = datetime.datetime(2021, 2, 28)
        with self.assertRaises(Exception):
            cepimose.vaccinations_date_range(
                end_date=end_date, start_date=start_date, property=property
            )

        start_date = datetime.datetime(2020, 12, 26)
        today = datetime.datetime.today()
        end_date = datetime.datetime(today.year, today.month, today.day)
        data = cepimose.vaccinations_date_range(
            end_date=end_date, start_date=start_date, property=property
        )
        self.assertDatesIncreaseSince(data.by_day, datetime.datetime(2020, 12, 26))

        self.assertEqual(start_date, data.date_from)
        self.assertEqual(end_date, data.date_to)

    def test_vaccinations_gender_by_date(self):
        # it takes to long to fetch data for whole period from 2020-12-27 until today

        test_date1 = datetime.datetime(2021, 1, 10)
        test_date2 = datetime.datetime(2021, 2, 10)
        test_date3 = datetime.datetime(2021, 3, 10)
        test_date4 = datetime.datetime(2021, 4, 10)
        test_date5 = datetime.datetime(2021, 5, 10)

        test_date6 = datetime.datetime(2021, 2, 7)
        test_date7 = datetime.datetime(2021, 4, 4)
        test_date8 = datetime.datetime(2021, 4, 5)

        data1 = cepimose.vaccinations_gender_by_date(test_date1)
        data2 = cepimose.vaccinations_gender_by_date(test_date2)
        data3 = cepimose.vaccinations_gender_by_date(test_date3)
        data4 = cepimose.vaccinations_gender_by_date(test_date4)
        data5 = cepimose.vaccinations_gender_by_date(test_date5)

        data6 = cepimose.vaccinations_gender_by_date(test_date6)
        data7 = cepimose.vaccinations_gender_by_date(test_date7)
        data8 = cepimose.vaccinations_gender_by_date(test_date8)

        def assertRow(row, expected_date, expected_data):
            print(row, expected_date, expected_data)
            self.assertEqual(row.date, expected_date)
            self.assertAlmostEqual(row.female_first, expected_data[0], delta=300)
            self.assertAlmostEqual(row.female_second, expected_data[1], delta=300)
            self.assertAlmostEqual(row.male_first, expected_data[2], delta=300)
            self.assertAlmostEqual(row.male_second, expected_data[3], delta=300)

        assertRow(data1, test_date1, [None, None, 1, 0])
        assertRow(data2, test_date2, [1536, 495, 916, 215])
        assertRow(data3, test_date3, [2851, 3024, 1870, 1902])
        assertRow(data4, test_date4, [4652, 76, 4400, 53])
        assertRow(data5, test_date5, [1009, 690, 1321, 681])

        assertRow(data6, test_date6, [None, None, None, None])
        assertRow(data7, test_date7, [835, 1, 749, 1])
        assertRow(data8, test_date8, [None, None, 1, 0])

    def test_vaccinations_gender_by_date_for_today(self):
        # not sure if this test is useful
        test_date_today = datetime.datetime.today()
        test_today_year = test_date_today.year
        test_today_month = test_date_today.month
        test_today_day = test_date_today.day
        test_today_without_time = datetime.datetime(
            test_today_year, test_today_month, test_today_day
        )
        data_today = cepimose.vaccinations_gender_by_date(test_today_without_time)
        print(f"Today: {test_date_today}")
        self.assertIsNot(data_today, None)



import datetime
from typing import Union
import requests
import time
from .data import (
    _source,
    _vaccinations_dashboard_headers,
    _lab_dashboard_headers,
    _vaccinations_by_day_req,
    _vaccinations_by_age_req,
    _vaccines_supplied_and_used_req,
    _vaccinations_by_region_req,
    _vaccines_supplied_by_manufacturer_req,
    _vaccines_supplied_by_manufacturer_cum_req,
    _vaccination_by_age_group_requests,
    _vaccinations_by_region_by_day_requests,
    _vaccinations_municipalities_share_req,
    _vaccinations_timestamp_req,
    _vaccinations_age_group_by_region_on_day_requests,
    _vaccination_by_manufacturer_supplied_used_requests,
    _vaccinations_gender_by_date_requests,
    _create_vaccinations_data_range_request,
    _vaccinations_by_manufacturer_used_request,
    _lab_start_ts_req,
    _lab_end_ts_req,
    _lab_PCR_tests_performed_req,
    _lab_PCR_total_tests_performed_req,
    _lab_active_cases_estimated_req,
    _lab_confirmed_total_male_req,
    _lab_total_vaccinated_first_dose_req,
    _lab_active_cases_100k_req,
    _lab_cases_total_confirmed_req,
    _lab_HAT_total_tests_performed_req,
    _lab_cases_confirmed_req,
    _lab_confirmed_total_female_req,
    _lab_total_vaccinated_fully_req,
    _lab_cases_avg_7Days_req,
    _lab_HAT_tests_performed_req,
)
from .parser import (
    _parse_vaccinations_by_age,
    _parse_vaccinations_by_day,
    _parse_vaccines_supplied_and_used,
    _parse_vaccinations_by_region,
    _parse_vaccines_supplied_by_manufacturer,
    _parse_vaccines_supplied_by_manufacturer_cum,
    _parse_vaccinations_by_age_group,
    _parse_vaccinations_by_region_by_day,
    _parse_vaccinations_by_municipalities_share,
    _parse_vaccinations_timestamp,
    _parse_vaccinations_age_group_by_region_on_day,
    _parse_vaccinations_by_manufacturer_supplied_used,
    _parse_vaccinations_gender_by_date,
    _parse_vaccinations_date_range,
    _parse_vaccinations_date_range_manufacturers_used,
    _create_vaccinations_by_manufacturer_parser,
    _parse_single_data,
)

from .types import (
    VaccinationByAgeRow,
    VaccinationByDayRow,
    VaccinationsByGender,
    VaccinationsDateRangeByGroup,
    VaccinationsDoses,
    VaccineSupplyUsage,
    VaccinationByRegionRow,
    VaccinationByManufacturerRow,
    VaccinationAgeGroupByRegionOnDay,
    LabDashboard,
)

from .enums import Manufacturer, Region, AgeGroup, Gender

DAY_DELTA = datetime.timedelta(days=1)
FIRST_DATE = datetime.datetime(2020, 12, 26)
TODAY_TIME = datetime.datetime.today()
TODAY = datetime.datetime(TODAY_TIME.year, TODAY_TIME.month, TODAY_TIME.day)


def _get_data(req, parse_response, headers=_vaccinations_dashboard_headers):
    resp = requests.post(_source, headers=headers, json=req)
    resp.raise_for_status()
    return parse_response(resp.json())


def vaccinations_timestamp():
    """Gets data refresh time

    Returns:
        datetime: datetime representing NIJZ data refresh time
    """
    return _get_data(_vaccinations_timestamp_req, _parse_vaccinations_timestamp)


def vaccinations_by_day() -> "list[VaccinationByDayRow]":
    """Gets number of vaccinations by day


    Returns:
        list: a list of VaccinationByDayRow representing number of vaccinated
        persons with first dose and fully vaccinated.
    """
    return _get_data(_vaccinations_by_day_req, _parse_vaccinations_by_day)


def vaccinations_by_age() -> "list[VaccinationByAgeRow]":
    """Gets number of vaccinations and shares per age group for today

    Returns:
        list: a list of VaccinationByAgeRow representing number of vaccinated
        persons with first dose, fully vaccinated and corresponding shares per
        age group for today
    """
    return _get_data(_vaccinations_by_age_req, _parse_vaccinations_by_age)


def vaccines_supplied_and_used() -> "list[VaccineSupplyUsage]":
    """Gets cumulative number of supplied and used vaccines by day

    Returns:
        list: a list of VaccineSupplyUsage representing total supplied and used
        vaccines on particular day
    """
    return _get_data(_vaccines_supplied_and_used_req, _parse_vaccines_supplied_and_used)


def vaccinations_by_region() -> "list[VaccinationByRegionRow]":
    """Gets number of vaccinations and shares per region for today

    Returns:
        list: a list of VaccinationByRegionRow representing number of vaccinated
        persons with first dose, fully vaccinated and corresponding shares per
        region for today
    """
    return _get_data(_vaccinations_by_region_req, _parse_vaccinations_by_region)


def vaccines_supplied_by_manufacturer() -> "list[VaccinationByManufacturerRow]":
    """Gets number of supplied vaccines by manufacturer by day

    Returns:
        list: a list of VaccinationByManufacturerRow representing supplied
        vaccines by manufacturer on particular day
    """
    return _get_data(
        _vaccines_supplied_by_manufacturer_req, _parse_vaccines_supplied_by_manufacturer
    )


def vaccines_supplied_by_manufacturer_cumulative() -> "list[VaccinationByManufacturerRow]":
    """Gets cumulative number of supplied vaccines by manufacturer by day

    Returns:
        list: a list of VaccinationByManufacturerRow representing supplied
        vaccines by manufacturer on particular day
    """
    return _get_data(
        _vaccines_supplied_by_manufacturer_cum_req,
        _parse_vaccines_supplied_by_manufacturer_cum,
    )


# by age group
def vaccinations_by_age_group(
    group: AgeGroup = None,
) -> "dict[AgeGroup,list[VaccinationByDayRow]] or list[VaccinationByDayRow]":
    """Gets cumulative number of vaccinations for age group by day

    Args:
        group (enum[AgeGroup] | None): represents age group, default is None


    Returns:
        dict [group = None]: a dict with keys AgeGroup enum members and values as
        list of VaccinationByDayRow respresenting  number of vaccinated
        persons with first dose and fully vaccinated.

        list [group = AgeGroup enum member]: a list of VaccinationByDayRow
        respresenting  number of vaccinated persons with first dose and fully
        vaccinated for enum member.

    """
    obj = {}
    if group == None:

        key_value = _vaccination_by_age_group_requests.items()
        for key, req_list in key_value:
            req = req_list[0]

            obj[key] = _get_data(req, _parse_vaccinations_by_age_group)
        return obj

    req = _vaccination_by_age_group_requests[group][0]
    return _get_data(req, _parse_vaccinations_by_age_group)


# by region by day
def vaccinations_by_region_by_day(
    region: Region = None,
) -> "dict[Region, list[VaccinationByDayRow]]":
    # TODO make same return if arg is not None as vaccinations_by_age_group and vaccinations_age_group_by_region_on_day or vice verse
    """Gets cumulative number of vaccinations for region by day

    Args:
        group (enum[Region] | None): represents region, default is None


    Returns:
        dict [group = None]: a dict with keys Region enum members and values as
        list of VaccinationByDayRow respresenting  number of vaccinated
        persons with first dose and fully vaccinated.
        If arg [group] is set then returns only for that group.
    """
    obj = {}
    if region == None:
        key_value = _vaccinations_by_region_by_day_requests.items()
        for key, req_list in key_value:
            req = req_list[0]
            doses = _get_data(req, _parse_vaccinations_by_region_by_day)
            obj[key] = doses
        return obj

    req = _vaccinations_by_region_by_day_requests[region][0]
    doses = _get_data(req, _parse_vaccinations_by_region_by_day)
    obj[region] = doses

    return obj


def vaccinations_age_group_by_region_on_day(
    group: AgeGroup = None,
) -> "Union[dict[AgeGroup, list[VaccinationAgeGroupByRegionOnDay]],list[VaccinationAgeGroupByRegionOnDay]]":
    """Gets number of vaccinations for regions in each age group

    Args:
        group (enum[AgeGroup] | None): represents age group, default is None

    Returns:
        dict [group = None]: a dict with keys AgeGroup enum members and values as
        list of VaccinationAgeGroupByRegionOnDay respresenting number of vaccinated persons with first dose and fully vaccinated per region as a list of VaccinationAgeGroupByRegionOnDayDose representing region, cumulative number of vaccinated persons in age group, number of vaccinated person in region for age group with coresponding shares.

        list [group = AgeGroup enum member]: a list of VaccinationAgeGroupByRegionOnDay respresenting number of vaccinated persons with first dose and fully vaccinated per region as a list of VaccinationAgeGroupByRegionOnDayDose representing region, cumulative number of vaccinated persons in age group, number of vaccinated person in region for age group with coresponding shares.
    """
    obj = {}

    if group == None:
        key_value = _vaccinations_age_group_by_region_on_day_requests.items()
        for key, req_list in key_value:
            req = req_list[0]
            doses = _get_data(req, _parse_vaccinations_age_group_by_region_on_day)
            obj[key] = doses
        return obj

    req = _vaccinations_age_group_by_region_on_day_requests[group][0]
    doses = _get_data(req, _parse_vaccinations_age_group_by_region_on_day)
    return doses


# PAGE 2
# municipalities
def vaccinations_by_municipalities_share():
    """Gets number of vaccinations per municipality on today.

    Returns:
        list: a list of VaccinationMunShare representing municipality, vaccinated persons with first dose and fully vaccinated with coresponding shares and municipality's population.

    """
    return _get_data(
        _vaccinations_municipalities_share_req,
        _parse_vaccinations_by_municipalities_share,
    )


# PAGE 3
# manufacturers
def vaccinations_by_manufacturer_supplied_used(
    group: Manufacturer = None,
) -> "dict[Manufacturer, list[VaccineSupplyUsage]] or list[VaccineSupplyUsage]":
    """Gets cumulative number of supplied and used vaccines on delivery date.

    Returns:
        dict [group = None]: a dict with keys Manufacturer enum members and values as
        list of VaccineSupplyUsage representing manufacturer's cumulative number of supplied and used vaccines

        list [group = Manufacturer enum member]: list of VaccineSupplyUsage representing manufacturer's cumulative number of supplied and used vaccines
    """
    obj = {}

    if group == None:
        key_value = _vaccination_by_manufacturer_supplied_used_requests.items()
        for key, req_list in key_value:
            req = req_list[0]
            doses = _get_data(req, _parse_vaccinations_by_manufacturer_supplied_used)
            obj[key] = doses
        return obj

    req = _vaccination_by_manufacturer_supplied_used_requests[group][0]
    doses = _get_data(req, _parse_vaccinations_by_manufacturer_supplied_used)
    return doses


def vaccinations_by_manufacturer_used() -> "list[VaccinationByManufacturerRow]":
    """Gets number of used vaccines per manufacturer per day.

    Warning: There is some discrepancy for first supply and first use of Moderna and Astra Zeneca vaccine! Both vaccines were used before first supply! See: 2021-01-08 for Moderna, 2021-01-28 for Astra Zeneca

    Returns:
        list: a list of VaccinationByManufacturerRow representing each manufacturer's number of used vaccines on day.
    """
    obj = {}
    for manu in Manufacturer:
        print(manu)
        manufacturer_parser = _create_vaccinations_by_manufacturer_parser(manu)
        obj[manu] = _get_data(
            _vaccinations_by_manufacturer_used_request[manu],
            manufacturer_parser,
        )

    start_date = FIRST_DATE + DAY_DELTA
    end_date = TODAY

    result = []
    while start_date <= end_date:
        pfizer = list(filter(lambda x: x.date == start_date, obj[Manufacturer.PFIZER]))
        moderna = list(
            filter(lambda x: x.date == start_date, obj[Manufacturer.MODERNA])
        )
        az = list(filter(lambda x: x.date == start_date, obj[Manufacturer.AZ]))
        janssen = list(
            filter(lambda x: x.date == start_date, obj[Manufacturer.JANSSEN])
        )
        try:
            pfizer_used = pfizer[0].dose if len(pfizer) != 0 else None
            moderna_used = moderna[0].dose if len(moderna) != 0 else None
            az_used = az[0].dose if len(az) != 0 else None
            janssen_used = janssen[0].dose if len(janssen) != 0 else None
            result.append(
                VaccinationByManufacturerRow(
                    start_date,
                    pfizer_used,
                    moderna_used,
                    az_used,
                    janssen_used,
                )
            )
        except:
            print(start_date, "Something went wrong")
            print(pfizer, moderna, az, janssen)
        start_date += DAY_DELTA
    return result


# PAGE 1
# gender
def vaccinations_gender_by_date(
    date: datetime.datetime = None,
) -> "Union[VaccinationsByGender, list[VaccinationsByGender]]":
    """Gets number of presons vaccinated with first dose and fully vaccinated persons by gender on day.

    Returns:
        list [date = None]: a list of VaccinationsByGender representing fully vaccinated and vaccinated persons with first dose by gender.

    """
    start = time.perf_counter()

    if date == None:
        result = []
        for day in _vaccinations_gender_by_date_requests:
            date = day["date"]
            female = day[Gender.FEMALE]
            male = day[Gender.MALE]
            female_first = _get_data(female[0], _parse_vaccinations_gender_by_date)
            female_second = _get_data(female[1], _parse_vaccinations_gender_by_date)
            male_first = _get_data(male[0], _parse_vaccinations_gender_by_date)
            male_second = _get_data(male[1], _parse_vaccinations_gender_by_date)
            result.append(
                VaccinationsByGender(
                    date=date,
                    female_first=female_first,
                    female_second=female_second,
                    male_first=male_first,
                    male_second=male_second,
                )
            )
        finish = time.perf_counter()
        print(f"Elapsed time: {finish - start}")
        return result

    filtered_days = list(
        filter(lambda item: item["date"] == date, _vaccinations_gender_by_date_requests)
    )

    if len(filtered_days) == 0:
        return None

    day = filtered_days[0]

    date = day["date"]
    female = day[Gender.FEMALE]
    male = day[Gender.MALE]
    female_first = _get_data(female[0], _parse_vaccinations_gender_by_date)
    female_second = _get_data(female[1], _parse_vaccinations_gender_by_date)
    male_first = _get_data(male[0], _parse_vaccinations_gender_by_date)
    male_second = _get_data(male[1], _parse_vaccinations_gender_by_date)
    finish = time.perf_counter()
    print(f"Elapsed time: {finish - start}")
    return VaccinationsByGender(
        date=date,
        female_first=female_first,
        female_second=female_second,
        male_first=male_first,
        male_second=male_second,
    )


# PAGE 1
def vaccinations_date_range(
    property: Union[Region, AgeGroup, None] = None,
    start_date: datetime.datetime = FIRST_DATE + DAY_DELTA,
    end_date: datetime.datetime = TODAY,
) -> VaccinationsDateRangeByGroup:
    """Gets cumulative number of fully and first does vaccinated persons by day in date range and cumulative number of vaccinated persons by gender and cumulative number of used vaccines by manufacturer on last day in date range either for region, age group or whole country.

    Source: PAGE 1 at https://app.powerbi.com/view?r=eyJrIjoiYWQ3NGE1NTMtZWJkMi00NzZmLWFiNDItZDc5YjU5MGRkOGMyIiwidCI6ImFkMjQ1ZGFlLTQ0YTAtNGQ5NC04OTY3LTVjNjk5MGFmYTQ2MyIsImMiOjl9

    Args:
        property (enum[Region] | enum[AgeGroup] | None): represents how data is filtered, default is None
        start_date (datetime): represent begining of date range
        end_date (datetime): represent begining of date range

    Returns:
        VaccinationsDateRangeByGroup: a VaccinationsDateRangeByGroup representing date range, cumulative vaccinations by day in date range and cumulative vaccinations by gender and by manufacturer on last date in date range.
    """

    if end_date < start_date:
        raise Exception(
            f"Argument [end_date]: {end_date} should be greater or equal than [start_date]: {start_date}"
        )

    if (
        not isinstance(property, Region)
        and not isinstance(property, AgeGroup)
        and property != None
    ):
        raise Exception(
            f"Argument [property] must be instance of Region, AgeGroup or None"
        )

    req = _create_vaccinations_data_range_request(
        end_date=end_date + DAY_DELTA, start_date=start_date, property=property
    )

    group = _get_data(req.group, _parse_vaccinations_date_range)

    result = VaccinationsDateRangeByGroup(start_date, end_date, property, group)

    male1 = _get_data(req.male1, _parse_vaccinations_gender_by_date)
    male2 = _get_data(req.male2, _parse_vaccinations_gender_by_date)
    result.male = VaccinationsDoses(male1, male2)

    female1 = _get_data(req.female1, _parse_vaccinations_gender_by_date)
    female2 = _get_data(req.female2, _parse_vaccinations_gender_by_date)
    result.female = VaccinationsDoses(female1, female2)

    manufacturers = _get_data(
        req.manufacturers, _parse_vaccinations_date_range_manufacturers_used
    )

    for manu in manufacturers:
        result.__setattr__(manu.name, VaccinationsDoses(manu.dose1, manu.dose2))

    return result


# DASHBOARD LAB


def lab_start_timestamp() -> datetime.datetime:
    """Gets NIJZ first COVID-19 case time

    Returns:
        datetime: datetime representing NIJZ first COVID-19 case
    """
    return _get_data(
        _lab_start_ts_req, _parse_vaccinations_timestamp, _lab_dashboard_headers
    )


def lab_end_timestamp() -> datetime.datetime:
    """Gets NIJZ last data update time

    Returns:
        datetime: datetime representing NIJZ last data update time
    """
    return _get_data(
        _lab_end_ts_req, _parse_vaccinations_timestamp, _lab_dashboard_headers
    )


def lab_PCR_tests_performed() -> int:
    """Gets performed PCR tests on today

    today -> date from lab_end_timestamp()

    Returns:
        int: a int number representing performed PCR tests on today
    """
    return _get_data(
        _lab_PCR_tests_performed_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_PCR_total_tests_performed() -> int:
    """Gets performed PCR tests in date range

    start -> date from lab_start_timestamp()
    end -> date from lab_end_timestamp()

    Returns:
        int: a int number representing performed PCR tests in date range
    """
    return _get_data(
        _lab_PCR_total_tests_performed_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_active_cases_estimated() -> int:
    """Gets estimated active cases

    Returns:
        int: a int number representing estimated active cases
    """
    return _get_data(
        _lab_active_cases_estimated_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_confirmed_total_male() -> int:
    """Gets male total confirmed cases in date range

    start -> date from lab_start_timestamp()
    end -> date from lab_end_timestamp()

    Returns:
        int: a int number representing male total confirmed cases in date range
    """
    return _get_data(
        _lab_confirmed_total_male_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_total_vaccinated_first_dose() -> int:
    """Gets vaccinated population with first dose in date range

    start -> date from lab_start_timestamp()
    end -> date from lab_end_timestamp()

    Returns:
        int: a int number representing vaccinated population with first dose in date range
    """
    return _get_data(
        _lab_total_vaccinated_first_dose_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_active_cases_100k() -> float:
    """Gets number of active cases per 100k population

    Returns:
        float: a float number representing active cases per 100k population
    """
    parsed_resp = _get_data(
        _lab_active_cases_100k_req, _parse_single_data, _lab_dashboard_headers
    )
    return float(parsed_resp)


def lab_cases_total_confirmed() -> int:
    """Gets total confirmed cases in date range

    start -> date from lab_start_timestamp()
    end -> date from lab_end_timestamp()

    Returns:
        int: a int number representing total confirmed cases in date range
    """
    return _get_data(
        _lab_cases_total_confirmed_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_HAT_total_tests_performed() -> int:
    """Gets performed HAT tests in date range

    start -> date from lab_start_timestamp()
    end -> date from lab_end_timestamp()

    Returns:
        int: a int number representing performed HAT tests in date range
    """
    return _get_data(
        _lab_HAT_total_tests_performed_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_cases_confirmed() -> int:
    """Gets confirmed cases on today

    today -> date from lab_end_timestamp()

    Returns:
        int: a int number representing confirmed cases on today
    """
    return _get_data(
        _lab_cases_confirmed_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_confirmed_total_female() -> int:
    """Gets female total confirmed cases in date range

    start -> date from lab_start_timestamp()
    end -> date from lab_end_timestamp()

    Returns:
        int: a int number representing female total confirmed cases in date range
    """
    return _get_data(
        _lab_confirmed_total_female_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_total_vaccinated_fully() -> int:
    """Gets fully vaccinated population in date range

    start -> date from lab_start_timestamp()
    end -> date from lab_end_timestamp()

    Returns:
        int: a int number representing fully vaccinated population in date range
    """
    return _get_data(
        _lab_total_vaccinated_fully_req, _parse_single_data, _lab_dashboard_headers
    )


def lab_cases_avg_7Days() -> float:
    """Gets 7 days average confirmed cases

    Returns:
        float: a float number representing 7 days average confirmed cases
    """
    parsed_resp = _get_data(
        _lab_cases_avg_7Days_req, _parse_single_data, _lab_dashboard_headers
    )
    return float(parsed_resp)


def lab_HAT_tests_performed():
    """Gets performed HAT tests on today

    today -> date from lab_end_timestamp()

    Returns:
        int: a int number representing performed HAT tests on today
    """
    return _get_data(
        _lab_HAT_tests_performed_req, _parse_single_data, _lab_dashboard_headers
    )


def get_lab_dashboard() -> LabDashboard:
    """Gets NIJZ dashboard:
        'Prikaz števila opravljenih cepljenj, testiranj in potrjenih okužb s covid-19 v SLoveniji'

    source: https://app.powerbi.com/view?r=eyJrIjoiMDc3MDk4MmQtOGE4NS00YTRkLTgyYjktNWQzMjk5ODNlNjVhIiwidCI6ImFkMjQ1ZGFlLTQ0YTAtNGQ5NC04OTY3LTVjNjk5MGFmYTQ2MyIsImMiOjl9&pageName=ReportSection24198f7e6d06db643832

    Returns:
        LabDashboard: a LabDashboard represent all data from dashboard
    """
    start_date = lab_start_timestamp()
    end_date = lab_end_timestamp()
    pcr = lab_PCR_tests_performed()
    pcr_total = lab_PCR_total_tests_performed()
    hat = lab_HAT_tests_performed()
    hat_total = lab_HAT_total_tests_performed()
    estimated_cases = lab_active_cases_estimated()
    cases_avg_100k = lab_active_cases_100k()
    cases_avg_7days = lab_cases_avg_7Days()
    cases = lab_cases_confirmed()
    cases_total = lab_cases_total_confirmed()
    cases_total_male = lab_confirmed_total_male()
    cases_total_female = lab_confirmed_total_female()
    vaccinated_first_dose = lab_total_vaccinated_first_dose()
    vaccinated_fully = lab_total_vaccinated_fully()

    result = LabDashboard(
        date=end_date,
        pcr=pcr,
        hat=hat,
        confirmed=cases,
        active_estimated=estimated_cases,
        cases_active_100k=cases_avg_100k,
        cases_active_7days=cases_avg_7days,
        date_start=start_date,
        pcr_total=pcr_total,
        hat_total=hat_total,
        confirmed_total=cases_total,
        male_total=cases_total_male,
        female_total=cases_total_female,
        vaccinated_first_dose=vaccinated_first_dose,
        vaccinated_fully=vaccinated_fully,
    )

    return result



import datetime
from typing import Union

from cepimose.enums import AgeGroup, Gender, Region, Manufacturer
from cepimose.types import CommandQueryFrom, DateRangeCommands_Requests


_ExecutionMetrics = {"ExecutionMetricsKind": 1}


def _get_Binding(projections: list, data_volume: int, primary: dict, version: int):
    return {
        "Primary": {"Groupings": [{"Projections": projections}]},
        "DataReduction": {"DataVolume": data_volume, "Primary": primary},
        "Version": version,
    }


def _get_Version(version: int = 2) -> "int":
    return version


def _get_Where(
    functions: list, arguments: list, special_args: list, skip_function: list
) -> "list":
    result = []
    for index, func in enumerate(functions):
        if not index in skip_function:
            result.append(func(*arguments[index], *special_args[index]))
    return result


# DATE RANGE GROUP QUERY FROM
def _get_From(*args: CommandQueryFrom) -> "list[dict]":
    result = []
    for arg in args:
        result.append({"Name": arg.name, "Entity": arg.entity, "Type": arg.type})
    return result


# DATE RANGE GROUP QUERY WHERE
def _get_Column(
    source: str,
    Property="Date",
):
    return {"Expression": {"SourceRef": {"Source": source}}, "Property": Property}


def _get_DateSpan(value=datetime.datetime(2020, 12, 26), time_unit=5):
    return {
        "Expression": {"Literal": {"Value": f"datetime'{value.isoformat()}'"}},
        "TimeUnit": time_unit,
    }


def _get_Comparison_With_DateSpan(comparison_kind, source, Property, value, time_unit):
    return {
        "ComparisonKind": comparison_kind,
        "Left": {"Column": _get_Column(source, Property)},
        "Right": {"DateSpan": _get_DateSpan(value, time_unit)},
    }


def _get_Comparison_With_Literal(
    comparison_kind: int,
    source,
    Property="Date",
    value=datetime.datetime(2020, 12, 26),
):
    Value = (
        f"datetime'{value.isoformat()}'"
        if isinstance(value, datetime.datetime)
        else value
    )

    return {
        "ComparisonKind": comparison_kind,
        "Left": {"Column": _get_Column(source, Property)},
        "Right": {"Literal": {"Value": Value}},
    }


def _get_Condition_Comparison_With_DateSpan(
    source: str,
    comparison_kind=1,
    Property="Date",
    value=datetime.datetime(2020, 12, 26),
    time_unit=5,
):
    return {
        "Condition": {
            "Comparison": _get_Comparison_With_DateSpan(
                comparison_kind, source, Property, value, time_unit
            )
        }
    }


def _get_Condition_Left_And_Right_Comparison(
    source: str, start_date: datetime.datetime, end_date: datetime.datetime
):
    left = {
        "Left": {
            "Comparison": _get_Comparison_With_Literal(2, source, "Date", start_date)
        }
    }
    right = {
        "Right": {
            "Comparison": _get_Comparison_With_Literal(3, source, "Date", end_date)
        },
    }
    return {"Condition": {"And": {**left, **right}}}


def _get_Condition_In_Expression(Property: str, source: str, value: str):
    return {
        "Condition": {
            "In": {
                "Expressions": [{"Column": _get_Column(source, Property)}],
                "Values": [[{"Literal": {"Value": f"'{value}'"}}]],
            }
        }
    }


# DATE RANGE GROUP QUERY SELECT
def _get_Group_Select(*options: list):
    def create_Select_item(item):
        return {item[0]: _get_Column(item[1], item[2]), "Name": f"{item[3]}.{item[2]}"}

    return list(map(create_Select_item, options))


_Date_Range_Group_Used_By_Day_Query_Options = {
    "common": {
        "Query": {
            "Version": _get_Version,
            "From": _get_From,
            "Where": [
                _get_Condition_Left_And_Right_Comparison,
                _get_Condition_In_Expression,
                _get_Condition_Comparison_With_DateSpan,
            ],
            "Select": _get_Group_Select,
        },
        "Binding": _get_Binding,
    },
    "args": {
        "Version": [],
        "From": [],
        "Where": [],
        "Select": [
            ["Column", "c1", "Date", "Calendar"],
            ["Measure", "c", "Weight running total in Date", "eRCO_podatki"],
            [
                "Measure",
                "c",
                "Tekoča vsota za mero Precepljenost v polju Date",
                "eRCO_podatki_ed",
            ],
        ],
        "OrderBy": [],
        "Binding": [[0, 1, 2], 4, {"BinnedLineSample": {}}, 1],
    },
    None: {
        "Where": [["c1"], [], ["c1"]],
        "From": [
            CommandQueryFrom("c", "eRCO_​​podatki", 0),
            CommandQueryFrom("c1", "Calendar", 0),
        ],
        "skip_common_where_index": [1],
    },
    Region: {
        "Where": [["c1"], ["Regija", "s"], ["c1"]],
        "From": [
            CommandQueryFrom("c", "eRCO_​​podatki", 0),
            CommandQueryFrom("c1", "Calendar", 0),
            CommandQueryFrom("s", "Sifrant_regija", 0),
        ],
    },
    AgeGroup: {
        "Where": [["c1"], ["Starostni ​razred", "x"], ["c1"]],
        "From": [
            CommandQueryFrom("c", "eRCO_​​podatki", 0),
            CommandQueryFrom("c1", "Calendar", 0),
            CommandQueryFrom("x", "xls_SURS_starost", 0),
        ],
    },
}


# GENDER
def _get_gender_Query_Select_Dose(dose):
    select = {
        "dose1": [
            {
                "Measure": _get_Column("e", "Weight for 1"),
                "Name": "eRCO_podatki.Weight for 1",
            }
        ],
        "dose2": [
            {
                "Aggregation": {
                    "Expression": {"Column": _get_Column("e", "Precepljenost")},
                    "Function": 0,
                },
                "Name": "Sum(eRCO_podatki_ed.Precepljenost)",
            }
        ],
    }
    return select[dose]


def _get_OrderBy(options: list = [0], select: "list[dict]" = []):
    if len(select) == 0:
        raise Exception("Empty arg [select]!")

    index = options[0]
    expression = {**select[index]}
    del expression["Name"]
    return [{"Direction": 2, "Expression": expression}]


_Date_Range_Group_Gender_Query_Options = {
    "common": {
        "Query": {
            "Version": _get_Version,
            "From": _get_From,
            "Where": [
                _get_Condition_In_Expression,
                _get_Condition_Left_And_Right_Comparison,
                _get_Condition_In_Expression,
                _get_Condition_Comparison_With_DateSpan,  # arg = "c"
            ],
            "Select": _get_gender_Query_Select_Dose,
            "OrderBy": _get_OrderBy,
        },
        "Binding": _get_Binding,
    },
    "args": {
        "Version": [],
        "From": [],
        "Where": [],
        "Select": [],
        "OrderBy": [[0]],
        "Binding": [[0], 3, {"Window": {}}, 1],
    },
    "specific": ["dose1", "dose2"],
    None: {
        "Where": [[], ["c"], [], ["c"]],
        "From": [
            CommandQueryFrom("e", "eRCO_​​podatki", 0),
            CommandQueryFrom("c", "Calendar", 0),
        ],
        "skip_common_where_index": [2],
    },
    Region: {
        "Where": [[], ["c"], ["Regija", "s"], ["c"]],
        "From": [
            CommandQueryFrom("c", "Calendar", 0),
            CommandQueryFrom("e", "eRCO_​​podatki", 0),
            CommandQueryFrom("s", "Sifrant_regija", 0),
        ],
    },
    AgeGroup: {
        "Where": [[], ["c"], ["Starostni ​razred", "x"], ["c"]],
        "From": [
            CommandQueryFrom("c", "Calendar", 0),
            CommandQueryFrom("e", "eRCO_​​podatki", 0),
            CommandQueryFrom("x", "xls_SURS_starost", 0),
        ],
    },
}


def _get_Condition_Not_Expression():
    return {
        "Condition": {
            "Not": {
                "Expression": {
                    "In": {
                        "Expressions": [{"Column": _get_Column("s", "Cepivo_Ime")}],
                        "Values": [[{"Literal": {"Value": "null"}}]],
                    }
                }
            }
        }
    }


_Date_Range_Group_Manufacturers_Query_Options = {
    "common": {
        "Query": {
            "Version": _get_Version,
            "From": _get_From,
            "Where": [
                _get_Condition_Not_Expression,
                _get_Condition_Left_And_Right_Comparison,
                _get_Condition_In_Expression,
                _get_Condition_Comparison_With_DateSpan,
            ],
            "Select": _get_Group_Select,
            "OrderBy": _get_OrderBy,
        },
        "Binding": _get_Binding,
    },
    "args": {
        "Version": [],
        "From": [],
        "Where": [],
        "Select": [
            ["Measure", "e", "Weight for 1", "eRCO_podatki"],
            ["Measure", "e", "Weight for 2", "eRCO_podatki"],
            ["Column", "s", "Cepivo_Ime", "Sifrant_Cepivo"],
        ],
        "OrderBy": [[1]],
        "Binding": [[0, 1, 2], 3, {"Window": {}}, 1],
    },
    None: {
        "Where": [[], ["c"], [], ["c"]],
        "From": [
            CommandQueryFrom("e", "eRCO_​​podatki", 0),
            CommandQueryFrom("s", "Sifrant_Cepivo", 0),
            CommandQueryFrom("c", "Calendar", 0),
        ],
        "skip_common_where_index": [2],
    },
    Region: {
        "Where": [[], ["c"], ["Regija", "s1"], ["c"]],
        "From": [
            CommandQueryFrom("e", "eRCO_​​podatki", 0),
            CommandQueryFrom("s", "Sifrant_Cepivo", 0),
            CommandQueryFrom("c", "Calendar", 0),
            CommandQueryFrom("s1", "Sifrant_regija", 0),
        ],
    },
    AgeGroup: {
        "Where": [[], ["c"], ["Starostni ​razred", "x1"], ["c"]],
        "From": [
            CommandQueryFrom("e", "eRCO_​​podatki", 0),
            CommandQueryFrom("s", "Sifrant_Cepivo", 0),
            CommandQueryFrom("c", "Calendar", 0),
            CommandQueryFrom("x1", "xls_SURS_starost", 0),
        ],
    },
}

_Date_Range_Group_Query_Options = {
    "by_day": _Date_Range_Group_Used_By_Day_Query_Options,
    "gender": _Date_Range_Group_Gender_Query_Options,
    "manufacturers": _Date_Range_Group_Manufacturers_Query_Options,
}


def _get_date_range_group_Query(
    common_options: dict,
    group_options: dict,
    version_args: list = [],
    where_args: list = [],
    select_args: list = [],
    order_by_args: list = [],
):

    from_args = group_options["From"]

    Query = common_options["Query"]

    skip_where_function_index = group_options.get("skip_common_where_index", [])

    select = Query["Select"](*select_args)

    where = _get_Where(
        Query["Where"], group_options["Where"], where_args, skip_where_function_index
    )

    order_by = Query.get("OrderBy", None)

    order_by = {"OrderBy": order_by(*order_by_args, select)} if order_by != None else {}
    command = {}
    command["Query"] = {
        "Version": Query["Version"](*version_args),
        "From": Query["From"](*from_args),
        "Select": select,
        "Where": where,
        **order_by,
    }

    return command


def _get_command(
    group: Union[Region, AgeGroup],
    command_type: str,
    where_args: list = [],
    specific_args: dict = {"Select": [], "OrderBy": []},
):
    command_options = _Date_Range_Group_Query_Options[command_type]
    common_options = command_options["common"]
    args_options = command_options["args"]

    version_args = args_options["Version"]
    select_args = [*specific_args["Select"], *args_options["Select"]]
    order_by_args = args_options["OrderBy"]
    binding_args = args_options["Binding"]

    command = {}
    command["SemanticQueryDataShapeCommand"] = {
        "Binding": common_options["Binding"](*binding_args),
        **_ExecutionMetrics,
    }
    group_type = type(group)
    group_options = command_options[group_type if group_type != type(None) else None]

    query = _get_date_range_group_Query(
        common_options,
        group_options,
        version_args,
        where_args,
        select_args,
        order_by_args,
    )

    command["SemanticQueryDataShapeCommand"] = {
        **query,
        **command["SemanticQueryDataShapeCommand"],
    }
    return command


def _get_gender_commands(
    start_date: datetime.datetime,
    end_date: datetime.datetime,
    group: Union[Region, AgeGroup, None],
):
    specific_options = _Date_Range_Group_Gender_Query_Options["specific"]

    commands = {}
    for gender in Gender:
        where_args = [
            ["OsebaSpol", "e", gender.value],
            [start_date, end_date],
            [group.value] if group != None else [],
            [],
        ]
        commands[gender] = [
            _get_command(
                group,
                "gender",
                where_args=where_args,
                specific_args={"Select": [dose]},
            )
            for dose in specific_options
        ]

    return commands


def _get_date_range_group_commands(
    start_date: datetime.datetime,
    end_date: datetime.datetime,
    group: Union[Region, AgeGroup, None],
) -> DateRangeCommands_Requests:

    where_by_day_args = [
        [start_date, end_date],
        [group.value] if group != None else [],
        [],
    ]
    where_manufacturers_args = [
        [],
        [start_date, end_date],
        [group.value] if group != None else [],
        [],
    ]

    by_day = _get_command(group, "by_day", where_by_day_args)

    gender = _get_gender_commands(start_date, end_date, group)
    male = gender[Gender.MALE]
    female = gender[Gender.FEMALE]

    manufacturers = _get_command(group, "manufacturers", where_manufacturers_args)

    commands = DateRangeCommands_Requests(by_day, *male, *female, manufacturers)

    return commands


def _get_default_manufacturer_used_command(manu: Manufacturer):
    return {
        "SemanticQueryDataShapeCommand": {
            "Query": {
                "Version": _get_Version(),
                "From": _get_From(
                    CommandQueryFrom("c1", "Calendar", 0),
                    CommandQueryFrom("s", "Sifrant_Cepivo", 0),
                    CommandQueryFrom("c", "eRCO_​​podatki", 0),
                ),
                "Select": [
                    {
                        "Column": _get_Column("c1", "Date"),
                        "Name": "Calendar.Date",
                    },
                    {
                        "Column": _get_Column("s", "Cepivo_Ime"),
                        "Name": "Sifrant_Cepivo.Cepivo_Ime",
                    },
                    {
                        "Aggregation": {
                            "Expression": {"Column": _get_Column("c", "Weight")},
                            "Function": 0,
                        },
                        "Name": "Sum(eRCO_podatki_ed.Weight)",
                    },
                ],
                "Where": [
                    _get_Condition_Comparison_With_DateSpan("c1", 2),
                    _get_Condition_Not_Expression(),
                    _get_Condition_In_Expression("Cepivo_Ime", "s", manu.value),
                ],
            },
            "Binding": {
                "Primary": {"Groupings": [{"Projections": [0, 2]}]},
                "Secondary": {"Groupings": [{"Projections": [1]}]},
                "DataReduction": {
                    "DataVolume": 4,
                    "Primary": {"Sample": {}},
                    "Secondary": {"Top": {}},
                },
                "Version": 1,
            },
            **_ExecutionMetrics,
        }
    }


def _create_manufacturers_used_commands():
    obj = {}
    for manu in Manufacturer:
        obj[manu] = _get_default_manufacturer_used_command(manu)

    return obj